from Johnson import *
import numpy as np
import gurobipy as gp
from gurobipy import GRB

def getData(file):
    with open(file) as f:
        firstLine = f.readline()
        firstLine = [x.strip() for x in firstLine.split(',')]
        numOfPatients = int(firstLine[0])
        lines = f.readlines()[numOfPatients:]
        preferences = {k:[] for k in range(numOfPatients)}

        for line in lines:
            x = line.split(",")
            patient = int(x[1])
            donor = int(x[0])

            preferences[patient].append(donor)

        f.close()

    return preferences

def getAltruists(file):
    altruist = []

    with open(file) as f:
        next(f)
        for line in f:
            x = line.split(",")
            if int(x[6]) == 1:
                altruist.append(int(x[0]) - 1)

    return altruist

def generateSCCGraph(scc):
	sccGraph = {k:[] for k in range(21)}

	for c in scc:
	    for v in c:
	        for donor in preferences[v]:
	            if donor in c:
	                sccGraph[donor].append(v)
	return sccGraph

file = "C:\\Users\\molly\\OneDrive\\Documents\\year 3\\FYP\\dataset\\25.md"

preferences = getData(file)

g = Graph(len(preferences))

for patient in preferences:
    for donor in preferences[patient]:
        g.add_edge(patient, donor)

scc = g.computeSCC()
sccGraph = generateSCCGraph(scc)

cycles = runJohnsonAlgorithm(scc, sccGraph)

patientFile = "C:\\Users\\molly\\OneDrive\\Documents\\year 3\\FYP\\kidney\\MD-00001-00000025.dat"
altruists = getAltruists(patientFile)

cycless = []

for c in cycles:
    a = set(c).intersection(altruists)
    if ((len(a) == 0) and (len(c) <= 4)):
        cycless.append(c)
    if (len(a) == 1):
        cycless.append(c)

cyclePairs = []
for c in cycless:
	for i in c:
	    if i not in cyclePairs:
	        cyclePairs.append(i)
cyclePairs.sort()

A = np.zeros((len(cyclePairs), len(cycless)))

cycleNum = 0
for c in cycless:
	for patient in c:
		A[cyclePairs.index(patient)][cycleNum] = 1
	cycleNum = cycleNum + 1

b = np.ones((len(cyclePairs),), dtype=int)
c = np.zeros((len(cycless),), dtype=int)

for cycle in cycless:
	c[cycless.index(cycle)] = len(cycle) - 1

print(c)

# Create a new model
m = gp.Model("model")

# Create variables
x = m.addMVar(shape=len(cycless), vtype=GRB.BINARY, name="x")

# Set objective
m.setObjective(c @ x, GRB.MAXIMIZE)

# Add constraint: Ax <= b
m.addConstr(A @ x <= b, "c1")

# Optimize model
m.optimize()

for v in m.getVars():
    print('%s = %g' % (v.varName, v.x))


print(len(m.getVars()))

for v in m.getVars():
    if v.x == 1:
        print('%s = %g' % (v.varName, v.x))