import copy
import random
from decimal import *

def removePatients(preferences, removeList):
	for patient in removeList: del preferences[patient]

def removePatient(preferences, patient):
	del preferences[patient]

def getUnmatchedPatients(preferences):
	remove = []

	for patient in preferences:
		if not preferences[patient]:
			remove.append(patient)
	return remove

def removeExpiredDonor(preferences, donor):
	for patient in preferences:
		for match in preferences[patient]:
			if donor == match:
				preferences[patient].remove(match)

	remove = []

def getUnsuitableDonors(preferences):
	unsuitableDonors = []

	preferredDonors = sorted({x for v in preferences.values() for x in v})

	for donor in preferences:
		if donor not in preferredDonors:
			unsuitableDonors.append(donor)
	return unsuitableDonors

def getAltruists(file):
    altruist = []

    with open(file) as f:
        next(f)
        for line in f:
            x = line.split(",")
            if int(x[6]) == 1:
                altruist.append(int(x[0]) - 1)

    return altruist

def checkCycleLength(matching, maxLength, altruists):
	keys = list(matching.keys())
	values = list(matching.values())

	for key in keys:
		keyIndex = keys.index(key)
		valueIndex = values.index(key)
		diff = abs(keyIndex - valueIndex)

		if diff >= k:
			return False

	return True

def checkAltruists(matching, altruists):
	keys = list(matching.keys())
	values = list(matching.values())

	start = 0
	while start < len(matching):
		end = values.index(keys[start]) + 1
		cycle = values[start:end]
		a = 0
		for altruist in altruists:
			if altruist in cycle[start:end]:
				a += 1
		if a > 1:
			return False
		start = end

	return True

def updateSolution():
	cycleLength = checkCycleLength(matches, k, altruist)
	altruistCheck = checkAltruists(matches, altruist)

	global mostMatched
	global bestSolution
	global maxSens

	if priority == "sens":
		totalSens = 0
		for patient in solution:
			totalSens += sens[patient]
		if (bestSolution == {} or (totalSens > maxSens) or (totalSens == maxSens and patientsMatched > mostMatched)) and cycleLength and altruistCheck:
			mostMatched = patientsMatched
			maxSens = totalSens
			bestSolution = copy.deepcopy(solution)

	if priority == "maximumPatients":
		if ((patientsMatched > mostMatched) or bestSolution == {}) and cycleLength and altruistCheck:
			mostMatched = patientsMatched
			bestSolution = copy.deepcopy(solution)

	if priority == "stage":
		priorityMatched = 0
		for patient in solution:
			if stage[patient] == 5:
				priorityMatched += 1
		if ((priorityMatched > mostMatched) or bestSolution == {} or ((priorityMatched == mostMatched) and patientsMatched > mostMatched)) and cycleLength and altruistCheck:
			mostMatched = patientsMatched
			maxPriorityMatched = priorityMatched
			bestSolution = copy.deepcopy(solution)

fileNumber = input("Leave blank for standard, -A for altruist and -AC for altruist and compatible pairs")

file = "C:\\Users\\molly\\OneDrive\\Documents\\year 3\\FYP\\dataset\\25" + fileNumber + ".md"

with open(file) as f:
    firstLine = f.readline()
    firstLine = [x.strip() for x in firstLine.split(',')]
    numOfPatients = int(firstLine[0])
    lines = f.readlines()[numOfPatients:]
    patientHLA = {k:[] for k in range(numOfPatients)}

    for line in lines:
        x = line.split(",")
        patient = int(x[1])
        donor = int(x[0])
        HLA = int(x[2])

        patientHLA[patient].append((donor, HLA))

f.close()

patientFile = "C:\\Users\\molly\\OneDrive\\Documents\\year 3\\FYP\\dataset\\25.csv"

sens = []
altruist = []
runPoints = []
pAge = []
dAge = []

with open(patientFile) as f:
	next(f)
	for line in f:
		x = line.split(",")
		sens.append(float(x[4]))
		runPoints.append(int(x[7]))
		pAge.append(int(x[8]))
		dAge.append(int(x[9]))

preferencesMatchingScore = {k:[] for k in range(numOfPatients)}

for patient in patientHLA:
	for donor in patientHLA[patient]:
		ageDiff = abs(dAge[patient] - pAge[patient])
		if ageDiff < 20:
			ageScore = 3
		else: ageScore = 0
		score = (runPoints[patient] * 50) + (sens[patient] / 2) + donor[1] + ageScore
		preferencesMatchingScore[patient].append((score, donor[0]))

for patient in preferencesMatchingScore:
	preferencesMatchingScore[patient] = sorted(preferencesMatchingScore[patient], reverse=True)

preferences = {k:[] for k in range(numOfPatients)}

for patient in preferencesMatchingScore:
	for t in preferencesMatchingScore[patient]:
		preferences[patient].append(t[1])

print("Initial Preferences", preferences)

altruist = getAltruists(patientFile)
stage = {k:[] for k in range(numOfPatients)}

for patient in stage:
	if random.random() < 0.7:
		stage[patient] = 5
	else:
		stage[patient] = 4

priority = input("To choose the priority of the matching:\n press s for sensitisation, st for stage 5 or m for maximum patients: ")

if priority == "s":
	priority = "sens"
if priority == "m":
	priority = "maximumPatients"
if priority == "st":
	priority = "stage"

k = int(input("Enter the number of patients allowed in a cycle "))

while getUnmatchedPatients(preferences):
	remove = getUnmatchedPatients(preferences)
	removePatients(preferences, remove)

	for item in remove:
		removeExpiredDonor(preferences, item)

unsuitableDonors = getUnsuitableDonors(preferences)
removePatients(preferences, unsuitableDonors)

originalPreferences = copy.deepcopy(preferences)

solution = {}
matches = {}
startM = 0
m = 0
mostMatched = 0
maxPriorityMatched = 0
bestSolution = {}
cycle = 0

while m < numOfPatients:
	if preferences:
		if m in preferences:
			matches[m] = preferences[m][0] #assign patient m to its first choice kidney
			removePatient(preferences, m) #patient m has now been assigned a kidney, so remove from preferences list
			removeExpiredDonor(preferences, matches[m]) #remove the assigned kidney from the preference of others
			if getUnmatchedPatients(preferences): #if there are patients with no remaining preferences
				while getUnmatchedPatients(preferences):
					remove = getUnmatchedPatients(preferences)
					removePatients(preferences, remove)
					expiredDonors = []
					for patient in preferences:
						for donor in preferences[patient]:
							if donor not in preferences:
								if donor not in expiredDonors and donor not in matches:
									expiredDonors.append(donor)
					for item in expiredDonors:
						removeExpiredDonor(preferences, item)
		else:
			if m in matches:
				cycle = 1
				patientsMatched = len(solution)
				solution.update(matches)
				m = list(preferences.keys())[0]
				checkpointPreferences = copy.deepcopy(preferences)
				updateSolution()
				continue
			else:
				if cycle == 1:
					preferences = checkpointPreferences
					m = list(preferences.keys())[0]
					matches = copy.deepcopy(solution)
					continue
				else:
					startM = startM + 1
					m = startM
					matches = {}
					checkpointPreferences = {}
					preferences = copy.deepcopy(originalPreferences)
					continue
		m = matches[m]
	else:
		solution.update(matches)
		isSolution = []

		for match in matches:
			if match not in matches.values():
				isSolution.append(False)
			else:
				isSolution.append(True)

		if all(isSolution):
			solution = matches
			patientsMatched = len(solution)
			updateSolution()

		startM = startM + 1
		m = startM
		matches = {} #reset matches as current ones do not hold a solution
		preferences = copy.deepcopy(originalPreferences)
		checkpointPreferences = {}
		solution = {}
		cycle = 0
		continue

if priority == "stage":
	print("Stage 5 patients matched: ", maxPriorityMatched)

print("Patients matched: ", mostMatched)
print("Optimal solution: ", bestSolution)

if priority == "sens":
	for p in bestSolution:
		print("Patient: ", p, " Sensitisation: ", sens[p])
